#include "texture.h"

#include <stb/stb_image.h>
#include <stb/stb_image_write.h>

#include "resource.h"
#include "internal.h"

namespace kek {

static unsigned int allocTexture(unsigned char *data, int width, int height, float borderColor[4]) {
	GLuint id;
	glGenTextures(1, &id);
	glBindTexture(GL_TEXTURE_2D, id);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameterfv(GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, borderColor);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_SRGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, data);
	//glTexImage2DMultisample(GL_TEXTURE_2D_MULTISAMPLE, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, data);
	glGenerateMipmap(GL_TEXTURE_2D);
	return id;
}

Texture::Texture(std::string texturePath) {
	this->path = texturePath;

	int width, height, nrChannels;
	MemoryBuffer *buf = Resource::loadResource(texturePath);
	if(!buf) {
		std::cout << "Failed to load texture: " << texturePath << std::endl;
		throw std::exception();
	}

	unsigned char *data = stbi_load_from_memory((stbi_uc *) buf->buffer, buf->length, &width, &height, &nrChannels, 0);
	delete buf;

	float borderColor[] = { 1.0f, 0.0f, 1.0f, 1.0f };
	allocTexture(data, width, height, borderColor);

	stbi_image_free(data);
}

Texture::Texture(GLuint id) {
	this->id = id;
}

Texture::~Texture() {
	glDeleteTextures(1, &id);
}

void Texture::use(GLenum texture) {
	glActiveTexture(texture);
	glBindTexture(GL_TEXTURE_2D, id);
}

std::shared_ptr<Texture> Texture::load(std::string texturePath) {
	std::shared_ptr<Texture> ret;

	auto iter = kekData.loadedTextures.find(texturePath);
	if(iter != kekData.loadedTextures.end()) ret = iter->second.lock();

	if(!ret) {
		ret = std::make_shared<Texture>(texturePath);
		kekData.loadedTextures.emplace(texturePath, ret);
	}

	return ret;
}

std::shared_ptr<Texture> Texture::generateColor(glm::vec3 color) {
	unsigned char *data = (unsigned char *) malloc(3 * sizeof(unsigned char));
	data[0] = (unsigned char) (color.x * 255);
	data[1] = (unsigned char) (color.y * 255);
	data[2] = (unsigned char) (color.z * 255);
	float borderColor[4] = { color.x, color.y, color.z, 1.0f };
	GLuint tex = allocTexture(data, 1, 1, borderColor);
	free(data);
	return std::make_shared<Texture>(tex);
}

}
